@@ -31,12 +31,19 @@ import subprocess
 import traceback
 import errno
 import time
-import pipes
 import socket
 import shutil
+import shlex
+from typing import (List, Optional, Union, Tuple)
 
 import adtlog
 
+# When running installed, this is /usr/share/autopkgtest.
+# When running uninstalled, this is the source tree.
+# Either way, it has a lib subdirectory from which virtualization
+# backends can read extra files.
+PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
+
 progname = "<VirtSubproc>"
 devnull_read = open('/dev/null', 'rb')
 caller = __main__
@@ -194,6 +201,58 @@ def check_exec(argv, downp=False, outp=False, timeout=0, fail_on_stderr=True):
     return out
 
 
+def load_shell_script(path: str) -> str:
+    if not os.path.isabs(path):
+        path = os.path.join(PKGDATADIR, path)
+
+    with open(path) as reader:
+        return reader.read()
+
+
+def wait_booted(
+    exec_command: List[str],
+    timeout: int = 120,
+    short_command_timeout: int = 10,
+    long_command_timeout: int = 60,
+) -> None:
+    '''
+    Wait until a container or VM with an init system has sufficiently
+    booted to interact with it.
+
+    timeout: how long we will continue to poll the container for status
+
+    short_command_timeout: how long we will wait for a relatively quick command
+    to finish
+
+    long_command_timeout: how long we will wait for a slower command to finish
+    '''
+    await_boot = load_shell_script('lib/in-testbed/await-boot.sh')
+
+    while timeout > 0:
+        timeout -= 1
+        time.sleep(1)
+        (rc, out, _) = execute_timeout(
+            None, 10, exec_command + ['runlevel'],
+            stdout=subprocess.PIPE,
+        )
+        if rc != 0:
+            adtlog.debug('wait_booted: polling status failed, retrying...')
+            continue
+        out = out.strip()
+        if out.split()[-1].isdigit():
+            adtlog.debug('waiting for network')
+            check_exec(
+                exec_command + ['sh', '-ec', await_boot],
+                timeout=long_command_timeout,
+            )
+            return
+
+        adtlog.debug('wait_booted: runlevel "%s", retrying...' % out)
+
+    bomb('timed out waiting for testbed to start; '
+         'last runlevel "%s"' % out)
+
+
 class timeout:
     def __init__(self, secs, exit_msg=None):
         '''Context manager that times out after given number of seconds.
@@ -229,9 +288,35 @@ def get_unix_socket(path):
     return s
 
 
-def expect(sock, search_bytes, timeout_sec, description=None, echo=False):
-    adtlog.debug('expect: "%s"' % (search_bytes or b'<none>').decode())
-    what = '"%s"' % (description or search_bytes or 'data')
+def matches_expectation(
+    haystack: bytes,
+    needle: Union[None, bytes, str, Tuple[Union[bytes, str]]],
+) -> bool:
+    if needle is None:
+        return True
+    elif isinstance(needle, tuple):
+        for n in needle:
+            if matches_expectation(haystack, n):
+                return True
+        else:
+            return False
+    elif isinstance(needle, bytes):
+        return needle in haystack
+    elif isinstance(needle, str):
+        return needle.encode('utf-8') in haystack
+    else:
+        raise TypeError(repr(needle))
+
+
+def expect(
+    sock: socket.socket,
+    search: Union[None, bytes, str, Tuple[Union[bytes, str]]],
+    timeout_sec: int,
+    description: Optional[str] = None,
+    echo: bool = False,
+):
+    adtlog.debug('expect: %r' % (search,))
+    what = repr(description or search or 'data')
     out = b''
     with timeout(timeout_sec,
                  description and ('timed out waiting for %s' % what) or None):
@@ -243,7 +328,8 @@ def expect(sock, search_bytes, timeout_sec, description=None, echo=False):
             if echo:
                 sys.stderr.buffer.write(block)
             out += block
-            if search_bytes is None or search_bytes in out:
+
+            if matches_expectation(out, search):
                 adtlog.debug('expect: found "%s"' % what)
                 break
 
@@ -309,9 +395,31 @@ def cmd_revert(c, ce):
     return [downtmp]
 
 
+def reboot_testbed():
+    (systemd_check, out, err) = execute_timeout(
+        None, 10, auxverb + ['test', '-d', '/run/systemd/system'],
+        stdout=subprocess.PIPE, stderr=subprocess.PIPE
+    )
+    if systemd_check == 0:
+        reboot_cmd = ['systemd-run', '--no-block', '--quiet', 'sh', '-c', 'sleep 3; reboot']
+        adtlog.debug('rebooting with command: %s' % reboot_cmd)
+        status, out, err = execute_timeout(None, 30, auxverb + reboot_cmd)
+
+        if status == 0:
+            return
+        else:
+            adtlog.debug(
+                'Cannot reboot with %s, trying something else' % reboot_cmd)
+
+    reboot_cmd = ['sh', '-c', '(sleep 3; reboot) >/dev/null 2>&1 &']
+    adtlog.debug('rebooting with command: %s' % reboot_cmd)
+    status, out, err = execute_timeout(None, 30, auxverb + reboot_cmd)
+
+
 def cmd_reboot(c, ce):
     global downtmp
     cmdnumargs(c, ce, 0, 1)
+    wait_reboot_args = {}
     if not downtmp:
         bomb("`reboot' when not open")
     if 'reboot' not in caller.hook_capabilities():
@@ -324,12 +432,12 @@ def cmd_reboot(c, ce):
                 '  tar --warning=none --create --absolute-names '
                 '''    -f $d/autopkgtest-tmpdir.tar '%s'; '''
                 '  rm -f /run/autopkgtest-reboot-prepare-mark; '
-                '  exit 0; fi; done; exit 1''' % (directories, downtmp)],
+                '  exit 0; fi; done; exit 1' % (directories, downtmp)],
                downp=True, timeout=copy_timeout)
     adtlog.debug('cmd_reboot: saved current downtmp, rebooting')
 
     try:
-        caller.hook_prepare_reboot()
+        wait_reboot_args = caller.hook_prepare_reboot() or {}
     except AttributeError:
         pass
 
@@ -337,9 +445,9 @@ def cmd_reboot(c, ce):
     if len(c) > 1 and c[1] == 'prepare-only':
         adtlog.info('state saved, waiting for testbed to reboot...')
     else:
-        execute_timeout(None, 30, auxverb +
-                        ['sh', '-c', '(sleep 3; reboot) >/dev/null 2>&1 &'])
-    caller.hook_wait_reboot()
+        reboot_testbed()
+
+    caller.hook_wait_reboot(**wait_reboot_args)
 
     # restore downtmp
     check_exec(['sh', '-ec', 'for d in %s; do '
@@ -508,7 +616,7 @@ def copyupdown_internal(wh, sd, upp):
 
     deststdout = devnull_read
     srcstdin = devnull_read
-    remfileq = pipes.quote(sd[iremote])
+    remfileq = shlex.quote(sd[iremote])
     if not dirsp:
         rune = 'cat %s%s' % ('><'[upp], remfileq)
         if upp:
@@ -566,8 +674,9 @@ def copyupdown_internal(wh, sd, upp):
             status = subprocs[sdn].wait()
             if not (status == 0 or (sdn == 0 and status == -13)):
                 timeout_stop()
-                bomb("%s %s failed, status %d" %
-                     (wh, ['source', 'destination'][sdn], status))
+                adtlog.info("%s %s failed, status %d" %
+                            (wh, ['source', 'destination'][sdn], status))
+                raise FailedCmd(['copy-failed'])
         timeout_stop()
     except Timeout:
         for sdn in [1, 0]:
@@ -652,8 +761,8 @@ def command():
     print(' '.join(r))
 
 
-signal_list = [	signal.SIGHUP, signal.SIGTERM,
-                signal.SIGINT, signal.SIGPIPE]
+signal_list = [signal.SIGHUP, signal.SIGTERM,
+               signal.SIGINT, signal.SIGPIPE]
 
 
 def sethandlers(f):
@@ -683,13 +792,13 @@ def error_cleanup():
         except Quit as q:
             sys.stderr.write(q.m)
             sys.stderr.write('\n')
-        except:
+        except Exception:
             sys.stderr.write('Unexpected cleanup error:\n')
             traceback.print_exc()
             sys.stderr.write('\n')
         if not ok:
             sys.stderr.write('while cleaning up because of another error:\n')
-    except:
+    except Exception:
         pass
 
 
@@ -721,7 +830,7 @@ def mainloop():
             sys.stderr.write(q.m)
             sys.stderr.write('\n')
         sys.exit(q.ec)
-    except:
+    except Exception:
         error_cleanup()
         sys.stderr.write('Unexpected error:\n')
         traceback.print_exc()
